#!/usr/bin/env python3
# encoding: utf-8
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
# Copyright (C) 2024 Henrik tom Wörden <h.tomwoerden@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import json
import os
import tempfile

import pytest
from caosadvancedtools.table_json_conversion.table_generator import (
    ColumnType, XLSXTemplateGenerator)
from openpyxl import load_workbook


def rfp(*pathcomponents):
    """
    Return full path.
    Shorthand convenience function.
    """
    return os.path.join(os.path.dirname(__file__), *pathcomponents)


def test_generate_sheets_from_schema():
    # trivial case; we do not support this
    schema = {}
    generator = XLSXTemplateGenerator()
    with pytest.raises(ValueError, match="Inappropriate JSON schema:.*"):
        generator._generate_sheets_from_schema(schema)

    # top level must be RT with Properties
    schema = {
        "type": "string"
    }
    with pytest.raises(ValueError, match="Inappropriate JSON schema:.*"):
        generator._generate_sheets_from_schema(schema)

    # bad type
    schema = {
        "type": "object",
        "properties": {
            "Training": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "str",
                        "description": "The name of the Record to be created"
                    },
                }
            }
        }
    }
    with pytest.raises(ValueError,
                       match="Inappropriate JSON schema: The following part "
                       "should define an object.*"):
        generator._generate_sheets_from_schema(schema)

    # bad schema
    schema = {
        "type": "object",
        "properties": {
            "Training": {
                "type": "object"
            }
        }
    }
    with pytest.raises(ValueError,
                       match="Inappropriate JSON schema: The following part "
                       "should define an object.*"):
        generator._generate_sheets_from_schema(schema)

    # minimal case: one RT with one P
    schema = {
        "type": "object",
        "properties": {
            "Training": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                        "description": "The name of the Record to be created"
                    },
                }
            }
        }
    }
    sdef = generator._generate_sheets_from_schema(schema)
    assert "Training" in sdef
    tdef = sdef['Training']
    assert 'name' in tdef
    assert tdef['name'] == (ColumnType.SCALAR, ["Training", 'name'])

    # example case
    with open(rfp("model_schema.json")) as sfi:
        schema = json.load(sfi)
    with pytest.raises(ValueError, match="A foreign key definition is missing.*"):
        generator._generate_sheets_from_schema(schema)
    sdef = generator._generate_sheets_from_schema(schema,
                                                  foreign_keys={'Training': (['date', 'url'], {})})
    assert "Training" in sdef
    tdef = sdef['Training']
    assert tdef['date'] == (ColumnType.SCALAR, ["Training", 'date'])
    assert tdef['url'] == (ColumnType.SCALAR, ["Training", 'url'])
    assert tdef['supervisor.family_name'] == (ColumnType.SCALAR, ["Training", 'supervisor',
                                                                  'family_name'])
    assert tdef['supervisor.given_name'] == (ColumnType.SCALAR, ["Training", 'supervisor',
                                                                 'given_name'])
    assert tdef['supervisor.Organisation.name'] == (ColumnType.SCALAR, ["Training", 'supervisor',
                                                                        'Organisation', 'name'])
    assert tdef['supervisor.Organisation.Country'] == (ColumnType.SCALAR, ["Training", 'supervisor',
                                                                           'Organisation', 'Country'])
    assert tdef['duration'] == (ColumnType.SCALAR, ["Training", 'duration'])
    assert tdef['participants'] == (ColumnType.SCALAR, ["Training", 'participants'])
    assert tdef['subjects'] == (ColumnType.LIST, ["Training", 'subjects'])
    assert tdef['remote'] == (ColumnType.SCALAR, ["Training", 'remote'])
    cdef = sdef['Training.coach']
    assert cdef['coach.family_name'] == (ColumnType.SCALAR, ["Training", 'coach', 'family_name'])
    assert cdef['coach.given_name'] == (ColumnType.SCALAR, ["Training", 'coach', 'given_name'])
    assert cdef['coach.Organisation.name'] == (ColumnType.SCALAR, ["Training", 'coach',
                                                                   'Organisation', 'name'])
    assert cdef['coach.Organisation.Country'] == (ColumnType.SCALAR, ["Training", 'coach',
                                                                      'Organisation', 'Country'])
    assert cdef['date'] == (ColumnType.FOREIGN, ["Training", 'date'])
    assert cdef['url'] == (ColumnType.FOREIGN, ["Training", 'url'])


def test_get_foreign_keys():
    generator = XLSXTemplateGenerator()
    fkd = {"Training": ['a']}
    assert ['a'] == generator._get_foreign_keys(fkd, ['Training'])

    fkd = {"Training": (['a'], {})}
    assert ['a'] == generator._get_foreign_keys(fkd, ['Training'])

    fkd = {"Training": {'hallo'}}
    with pytest.raises(ValueError, match=r"A foreign key definition is missing for path:\n\[\]\n{'Training': \{'hallo'\}\}"):
        generator._get_foreign_keys(fkd, ['Training'])

    fkd = {"Training": (['a'], {'b': ['c']})}
    assert ['c'] == generator._get_foreign_keys(fkd, ['Training', 'b'])

    with pytest.raises(ValueError, match=r"A foreign key definition is missing for.*"):
        generator._get_foreign_keys({}, ['Training'])


def test_get_max_path_length():
    assert 4 == XLSXTemplateGenerator._get_max_path_length({'a': (1, [1, 2, 3]),
                                                            'b': (2, [1, 2, 3, 4])})


def test_template_generator():
    generator = XLSXTemplateGenerator()
    with open(rfp("model_schema.json")) as sfi:
        schema = json.load(sfi)
    path = os.path.join(tempfile.mkdtemp(), 'test.xlsx')
    assert not os.path.exists(path)
    generator.generate(schema=schema, foreign_keys={'Training': (['date', 'url'], {})}, filepath=path)
    assert os.path.exists(path)
    generated = load_workbook(path)  # workbook can be read
    example = load_workbook(rfp("example_template.xlsx"))
    assert generated.sheetnames == example.sheetnames
    for sheetname in example.sheetnames:
        gen_sheet = generated[sheetname]
        ex_sheet = example[sheetname]
        for irow, (erow, grow) in enumerate(zip(ex_sheet.iter_rows(), gen_sheet.iter_rows())):
            for icol, (ecol, gcol) in enumerate(zip(erow, grow)):
                cell = gen_sheet.cell(irow+1, icol+1)
                assert ecol.value == gcol.value, f"Sheet: {sheetname}, cell: {cell.coordinate}"

    # test some hidden
    ws = generated.active
    assert ws.row_dimensions[1].hidden is True
    assert ws.column_dimensions['A'].hidden is True





    rp = '/home/henrik/CaosDB/management/external/dimr/eingabemaske/django/laforms/settings/DevelSchema.json'
    with open(rp) as sfi:
        schema = json.load(sfi)
    generator.generate(schema=schema,
                       foreign_keys={},
                       filepath=path)
    os.system(f'libreoffice {path}')

    # TODO test colisions of sheet or colnames
    # TODO test escaping of values

    # TODO finish enum example
